==> Smdn.Net.Imap4.Client-1.04.tar.bz2: Smdn.Net.Imap4.Client-1.04/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Client/ImapMessageInfo.FetchMessageBodyStream.cs <==
//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
#if NET_4_0
using System.Threading;
using System.Threading.Tasks;
#endif

using Smdn.Formats.Mime;
using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  partial class ImapMessageInfo {
    /// <param name="section">section-part. must be validated.</param>
    internal Tuple<Stream, IImapMessageAttribute> OpenRead(string section,
                                                           ImapPartialRange? range,
                                                           string encoding,
                                                           ImapMessageFetchBodyOptions options,
                                                           ImapFetchDataItem fetchDataItem)
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      return PrepareStream(new FetchMessageBodyStream(this,
                                                      options,
                                                      section,
                                                      range),
                           fetchDataItem,
                           encoding);
    }

#if NET_4_0
    private Stream OpenFetchMessageBodyStream(ImapMessageFetchBodyOptions options,
                                              CancellationToken cancellationToken,
#if NET_4_5
                                              IProgress<ImapMessageFetchProgressInfo> progress)
#else
                                              object progress)
#endif
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      // TODO: check busy and lock session

      return PrepareStream(new FetchMessageBodyStream(this,
                                                      options,
                                                      null,
                                                      null,
                                                      cancellationToken,
                                                      progress),
                           null,
                           null).Item1;
    }

    private Task<TResult> OpenFetchMessageBodyStreamAsync<TResult>(ImapMessageFetchBodyOptions options,
                                                                   CancellationToken cancellationToken,
#if NET_4_5
                                                                   IProgress<ImapMessageFetchProgressInfo> progress,
#else
                                                                   object progress,
#endif
                                                                   Converter<Stream, TResult> converter)
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      // TODO: check busy and lock session

      return Task.Factory.StartNew(() => {
        var stream = PrepareStream(new FetchMessageBodyStream(this,
                                                              options,
                                                              null,
                                                              null,
                                                              cancellationToken,
                                                              progress),
                                   null,
                                   null).Item1;

        if (cancellationToken.IsCancellationRequested)
          stream.Close();

        cancellationToken.ThrowIfCancellationRequested();

        return converter(stream);
      }, cancellationToken);
    }
#endif

    private Tuple<Stream, IImapMessageAttribute> PrepareStream(FetchMessageBodyStream stream,
                                                               ImapFetchDataItem fetchDataItem,
                                                               string encoding)
    {
      IImapMessageAttribute attrs;

      Mailbox.ProcessResult(stream.Prepare(fetchDataItem, out attrs));

      if (encoding == null)
        return Tuple.Create<Stream, IImapMessageAttribute>(stream,
                                                           attrs);
      else
        return Tuple.Create<Stream, IImapMessageAttribute>(ContentTransferEncoding.CreateDecodingStream(stream, encoding),
                                                           attrs);
    }

    private Stream OpenReadHeaderFields(IEnumerable<string> fetchHeaderFields,
                                        ImapMessageFetchBodyOptions options)
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      var stream = new FetchMessageBodyStream(this, options, fetchHeaderFields);

      IImapMessageAttribute discard;

      Mailbox.ProcessResult(stream.Prepare(null, out discard));

      return stream;
    }

    private sealed class FetchMessageBodyStream : ImapFetchMessageBodyStream {
      public ImapMessageInfo Message {
        get; private set;
      }

      public FetchMessageBodyStream(ImapMessageInfo message,
                                    ImapMessageFetchBodyOptions options,
                                    string fetchSection,
                                    ImapPartialRange? fetchRange)
        : base(message.Client.Session,
               FetchBodyByPeek(options),
               message.SequenceOrUidSet,
               fetchSection,
               fetchRange,
               DefaultFetchBlockSize)
      {
        this.Message = message;
      }

#if NET_4_0
      public FetchMessageBodyStream(ImapMessageInfo message,
                                    ImapMessageFetchBodyOptions options,
                                    string fetchSection,
                                    ImapPartialRange? fetchRange,
                                    CancellationToken cancellationToken,
#if NET_4_5
                                    IProgress<ImapMessageFetchProgressInfo> progress)
#else
                                    object progress)
#endif
        : base(message.Client.Session,
               FetchBodyByPeek(options),
               message.SequenceOrUidSet,
               fetchSection,
               fetchRange,
               DefaultFetchBlockSize,
               cancellationToken)
      {
        this.Message = message;
#if NET_4_5
        this.progress = progress;
#endif
      }
#endif

      public FetchMessageBodyStream(ImapMessageInfo message,
                                    ImapMessageFetchBodyOptions options,
                                    IEnumerable<string> fetchHeaderFields)
        : base(message.Client.Session,
               FetchBodyByPeek(options),
               message.SequenceOrUidSet,
               fetchHeaderFields,
               DefaultFetchBlockSize)
      {
        this.Message = message;
      }

      private static bool FetchBodyByPeek(ImapMessageFetchBodyOptions options)
      {
        return (int)(options & ImapMessageFetchBodyOptions.SetSeen) == 0;
      }

#if NET_4_5
      internal protected override void ReportFetchedOctetCount(long fetchedOctetCount, long totalOctetCount)
      {
        if (progress != null)
          progress.Report(new ImapMessageFetchProgressInfo(fetchedOctetCount, totalOctetCount, CancellationToken));
      }
#endif

      protected override Exception GetNoSuchMessageException(ImapCommandResult result)
      {
        Message.Mailbox.ProcessResult(result);

        Message.SetDeleted();

        return new ImapMessageVanishedException(Message);
      }

      protected override Exception GetFetchFailureException(ImapCommandResult result)
      {
        Message.Mailbox.ProcessResult(result);

        return base.GetFetchFailureException(result);
      }

#if NET_4_5
      private readonly IProgress<ImapMessageFetchProgressInfo> progress;
#endif
    }
  }
}

==> Smdn.Net.Imap4.Client-1.09.tar.bz2: Smdn.Net.Imap4.Client-1.09/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Client/ImapMessageInfo.FetchMessageBodyStream.cs <==
//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
#if NET_4_0
using System.Threading;
using System.Threading.Tasks;
#endif

using Smdn.Formats.Mime;
using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  partial class ImapMessageInfo {
    /// <param name="section">section-part. must be validated.</param>
    internal Stream OpenRead(string section,
                             ImapPartialRange? range,
                             string encoding,
                             ImapMessageFetchBodyOptions options)
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      Client.SetBusy();

      IImapMessageAttribute discard;

      return PrepareStream(new FetchMessageBodyStream(this,
                                                      options,
                                                      section,
                                                      range,
                                                      true),
                           null,
                           encoding,
                           out discard);
    }

    internal Stream OpenReadWithAttributes(ImapFetchDataItem fetchDataItem,
                                           bool setBusy,
                                           out IImapMessageAttribute messageAttributes)
    {
      messageAttributes = null;

      CheckMessageNumberAvailable(SequenceOrUidSet);

      if (setBusy)
        Client.SetBusy();

      return PrepareStream(new FetchMessageBodyStream(this,
                                                      ImapMessageFetchBodyOptions.Default,
                                                      null,
                                                      null,
                                                      setBusy),
                           fetchDataItem,
                           null,
                           out messageAttributes);
    }

#if NET_4_0
    private Stream OpenFetchMessageBodyStream(ImapMessageFetchBodyOptions options,
                                              CancellationToken cancellationToken,
#if NET_4_5
                                              IProgress<ImapMessageFetchProgressInfo> progress)
#else
                                              object progress)
#endif
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      Client.SetBusy();

      IImapMessageAttribute discard;

      return PrepareStream(new FetchMessageBodyStream(this,
                                                      options,
                                                      null,
                                                      null,
                                                      cancellationToken,
                                                      progress),
                           null,
                           null,
                           out discard);
    }

    private Task<TResult> OpenFetchMessageBodyStreamAsync<TResult>(ImapMessageFetchBodyOptions options,
                                                                   CancellationToken cancellationToken,
#if NET_4_5
                                                                   IProgress<ImapMessageFetchProgressInfo> progress,
#else
                                                                   object progress,
#endif
                                                                   Converter<Stream, TResult> converter)
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      return Client.StartTaskExclusively(() => {
        IImapMessageAttribute discard;

        var stream = PrepareStream(new FetchMessageBodyStream(this,
                                                              options,
                                                              null,
                                                              null,
                                                              cancellationToken,
                                                              progress),
                                   null,
                                   null,
                                   out discard);

        if (cancellationToken.IsCancellationRequested)
          stream.Close();

        cancellationToken.ThrowIfCancellationRequested();

        return converter(stream);
      }, cancellationToken, false);
    }
#endif

    private Stream PrepareStream(FetchMessageBodyStream stream,
                                 ImapFetchDataItem fetchDataItem,
                                 string encoding,
                                 out IImapMessageAttribute attributes)
    {
      try {
        Mailbox.ProcessResult(stream.Prepare(fetchDataItem, out attributes));

        if (encoding == null)
          return stream;
        else
          return ContentTransferEncoding.CreateDecodingStream(stream, encoding);
      }
      catch {
        stream.Close();

        throw;
      }
    }

    private Stream OpenReadHeaderFields(IEnumerable<string> fetchHeaderFields,
                                        ImapMessageFetchBodyOptions options)
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      Client.SetBusy();

      var stream = new FetchMessageBodyStream(this, options, fetchHeaderFields);

      IImapMessageAttribute discard;

      Mailbox.ProcessResult(stream.Prepare(null, out discard));

      return stream;
    }

    private sealed class FetchMessageBodyStream : ImapFetchMessageBodyStream {
      public ImapMessageInfo Message {
        get; private set;
      }

      public FetchMessageBodyStream(ImapMessageInfo message,
                                    ImapMessageFetchBodyOptions options,
                                    string fetchSection,
                                    ImapPartialRange? fetchRange,
                                    bool unsetBusy)
        : base(message.Client.Session,
               FetchBodyByPeek(options),
               message.SequenceOrUidSet,
               fetchSection,
               fetchRange,
               DefaultFetchBlockSize)
      {
        this.Message = message;
        this.unsetBusy = unsetBusy;
      }

#if NET_4_0
      public FetchMessageBodyStream(ImapMessageInfo message,
                                    ImapMessageFetchBodyOptions options,
                                    string fetchSection,
                                    ImapPartialRange? fetchRange,
                                    CancellationToken cancellationToken,
#if NET_4_5
                                    IProgress<ImapMessageFetchProgressInfo> progress)
#else
                                    object progress)
#endif
        : base(message.Client.Session,
               FetchBodyByPeek(options),
               message.SequenceOrUidSet,
               fetchSection,
               fetchRange,
               DefaultFetchBlockSize,
               cancellationToken)
      {
        this.Message = message;
        this.unsetBusy = true;
#if NET_4_5
        this.progress = progress;
#endif
      }
#endif

      public FetchMessageBodyStream(ImapMessageInfo message,
                                    ImapMessageFetchBodyOptions options,
                                    IEnumerable<string> fetchHeaderFields)
        : base(message.Client.Session,
               FetchBodyByPeek(options),
               message.SequenceOrUidSet,
               fetchHeaderFields,
               DefaultFetchBlockSize)
      {
        this.Message = message;
        this.unsetBusy = true;
      }

      private static bool FetchBodyByPeek(ImapMessageFetchBodyOptions options)
      {
        return (int)(options & ImapMessageFetchBodyOptions.SetSeen) == 0;
      }

      protected override void Dispose(bool disposing)
      {
        if (Message != null) {
          if (unsetBusy)
            Message.Client.UnsetBusy();

          Message = null;
        }

        base.Dispose(disposing);
      }

      internal protected override ImapCommandResult Prepare(ImapFetchDataItem fetchDataItem,
                                                            out IImapMessageAttribute messageAttr)
      {
        CheckDisposed();

        try {
          return base.Prepare(fetchDataItem, out messageAttr);
        }
        catch {
          Close();

          throw;
        }
      }

#if NET_4_5
      internal protected override void ReportFetchedOctetCount(long fetchedOctetCount, long totalOctetCount)
      {
        if (progress != null)
          progress.Report(new ImapMessageFetchProgressInfo(fetchedOctetCount, totalOctetCount, CancellationToken));
      }
#endif

      protected override Exception GetNoSuchMessageException(ImapCommandResult result)
      {
        CheckDisposed();

        Message.Mailbox.ProcessResult(result);

        Message.SetDeleted();

        return new ImapMessageVanishedException(Message);
      }

      protected override Exception GetFetchFailureException(ImapCommandResult result)
      {
        CheckDisposed();

        Message.Mailbox.ProcessResult(result);

        return base.GetFetchFailureException(result);
      }

      private void CheckDisposed()
      {
        if (Message == null)
          throw new ObjectDisposedException(GetType().FullName);
      }

      private readonly bool unsetBusy = true;
#if NET_4_5
      private readonly IProgress<ImapMessageFetchProgressInfo> progress;
#endif
    }
  }
}
